/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>

#include "sambascanner.h"
#include "Tools.h"
#include "IP_Range.h"
#include "Processes.h"
#include "String.h"
#include "Parser.h"
#include "Mutex.h"
#include "Threads.h"

int GetUniqueThreadID() {
	static pthread_mutex_t LockUniqueID = PTHREAD_MUTEX_INITIALIZER;
	static int UniqueID = 0;
	int ReturnID;
	pthread_mutex_lock(&LockUniqueID);
	UniqueID++;
	ReturnID = UniqueID;
	pthread_mutex_unlock(&LockUniqueID);
	return ReturnID;
}

void* Redirect_Thread(void *Parameters) {
	struct Redirect_Thread_InputData *Input = (struct Redirect_Thread_InputData*)Parameters;
	char Buffer[BUFSIZE_SHORT];
	int Read;
	int Cnt = 0;

	pthread_mutex_lock(&Input->Active);

	int First = 1;
	do {
		Cnt++;
		int Error;
		errno = 0;
		Read = read(Input->Read_FD, Buffer, sizeof(Buffer));
		Error = errno;
		if (Read <= 0) message(4, "[TID %d] Read %d: %d %d %s (%d -> %d)\n", Input->UniqueThreadID, Cnt, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
		/* TODO: This error sometimes (rarely) happens - but why? */
		/*
		if (Error == EBADF) {
			printf("ERROR DETECTED\n");
			fflush(stdout);
			fflush(stderr);
			*((int*)0) = 1;
		}
		*/
		if ((First) && (Error == EBADF)) {
			/* Wait for FD to become valid */
			usleep(100 * 1000);
			continue;
		}
		if (Read > 0) {
			double Time;
			int Written;

			First = 0;
			errno = 0;
			Written = write(Input->Outfile_FD, Buffer, Read);
			Error = errno;
			if (Written <= 0) message(4, "[TID %d] Write %d: %d %d %s (%d -> %d)\n", Input->UniqueThreadID, Cnt, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
			Time = GetTime();
			(*Input->Processes)[Input->ProcessID].TermTime = Time + (Input->TermTime * Options.Timings);
			(*Input->Processes)[Input->ProcessID].KillTime = Time + ((Input->TermTime + Input->KillTime) * Options.Timings);
		}
	} while (Read > 0);
	message(4, "[TID %d] FD %d -> Redirect_Thread -> FD %d Quit.\n", Input->UniqueThreadID, Input->Read_FD, Input->Outfile_FD);
	close(Input->Read_FD);
	close(Input->Outfile_FD);

	free(Input);		/* Invalid write! */
	pthread_exit(NULL);
	return NULL;
}

void* Process_Watcher_Thread(void *Parameters) {
	struct Process_Watcher_Thread_InputData *Input = (struct Process_Watcher_Thread_InputData*)Parameters;
	int i;
	double Time;

	while (Input->Running) {
		Time = GetTime();
		for (i = 0; i < *(Input->NumberProcesses); i++) {
			if (((*Input->Processes)[i].Running == 1) && ((*Input->Processes)[i].Valid == 1)) {
				if (Time > (*Input->Processes)[i].KillTime) {
					message(3, "=> SIGKILL Processes[%d] [PID %d]\n", i, (*Input->Processes)[i].PID);
					(*Input->Processes)[i].Signal = SIGKILL;
					kill((*Input->Processes)[i].PID, SIGKILL);
				} else {
					if (Time > (*Input->Processes)[i].TermTime) {
						message(3, "=> SIGTERM Processes[%d] [PID %d]\n", i, (*Input->Processes)[i].PID);
						(*Input->Processes)[i].Signal = SIGTERM;
						kill((*Input->Processes)[i].PID, SIGTERM);
					}
				}
			}
		}
		usleep(100*1000);	/* 100ms */
	}
	Input->Running = -1;
	pthread_exit(NULL);
	return NULL;
}

void* Process_Host_Thread(void *Parameters) {
	struct Process_Host_Thread_InputData *Input = (struct Process_Host_Thread_InputData*)Parameters;
	char Buffer[BUFSIZE_SHORT];
	char FileBuffer[BUFSIZE_SHORT];
	int Result, ProcID;
	int fd;

	strcpy(FileBuffer, TEMPPATH "sambascanner_process_hostXXXXXX");
	errno = 0;
	fd = (int)mkstemp(FileBuffer);
	if (errno != 0) {
		message(1, "Couldn't create temporary file in Process_Host_Thread: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	message(2, "Processing host: '%s'\n", Input->Hostname);
	snprintf(Buffer, sizeof(Buffer), "-g -N -L //%s", Input->Hostname);
	ProcID = Exec_Timeout(Options.SmbClient_Executable, Buffer, 5.0, 2.0, fd, Input->Processes, Input->NumberProcesses);
	Result = WaitFor(ProcID, Input->Processes);

	pthread_mutex_lock(Input->NumberMutex);
	if (Result < 0) {
		/* Process was killed */
		(*Input->Hostlookup_Processes_Failed)++;
	} else {
		(*Input->Hostlookup_Processes_Succeeded)++;
	}
	pthread_mutex_unlock(Input->NumberMutex);

	Parse_Hostinfo(Input->Hostname, FileBuffer, Input->Hostlist, Input->Sharelist);
	if (!Input->PreserveTemp) unlink(FileBuffer);
	
	free(Input);
	ThreadCnt_Dec();
	pthread_exit(NULL);
	return NULL;
}

void* Process_Share_Thread(void *Parameters) {
	struct Process_Share_Thread_InputData *Input = (struct Process_Share_Thread_InputData*)Parameters;
	char CmdParameters[BUFSIZE_SHORT];
	char Buffer[BUFSIZE_SHORT];
	char Buffer2[BUFSIZE_SHORT];
	char FileBuffer[BUFSIZE_SHORT];
	int Result, ProcID;
	int fd;

	strcpy(FileBuffer, TEMPPATH "sambascanner_process_sharesXXXXXX");
	errno = 0;
	fd = (int)mkstemp(FileBuffer);
	if (errno != 0) {
		message(1, "Couldn't create temporary file in Process_Share_Thread: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	Add_String_Atomic(Input->Resultfilelist, FileBuffer);
	
	message(2, "Now revealing content of smb:%s with unique thread ID %d\n", Input->Sharename, Input->UniqueThreadID);

	if (!Input->TryPasswords) {
		CmdParameters[0] = 0;
	} else {
		strcpy(CmdParameters, "-g");
	}
	if (Options.UserPass.Entries > 0) {
		int i;
		for (i = 0; i < Options.UserPass.Entries; i++) {
			snprintf(Buffer, sizeof(Buffer), "%s -u %s", CmdParameters, Options.UserPass.Entry[i]);
			strcpy(CmdParameters, Buffer);
		}
	}
		
	Escape_Apostrophes(Buffer2, Input->Sharename, sizeof(Buffer2));
	snprintf(Buffer, sizeof(Buffer), "%s 'smb:%s'", CmdParameters, Buffer2);
	strcpy(CmdParameters, Buffer);

	ProcID = Exec_Timeout(Options.SmbLister_Executable, CmdParameters, 90.0, 5.0, fd, Input->Processes, Input->NumberProcesses);
	Result = WaitFor(ProcID, Input->Processes);
	
	ThreadCnt_Dec();
	pthread_exit(NULL);
	return NULL;
}

void *Process_IP_Address_Thread(void *Parameters) {
	struct Process_IP_Address_Thread_InputData *Input = (struct Process_IP_Address_Thread_InputData*)Parameters;
	char Buffer[BUFSIZE_SHORT];
	char FileBuffer[BUFSIZE_SHORT];
	int Result, ProcID;
	int fd;

	message(3, "Now processing IP %d.%d.%d.%d\n", IP_ADDRESS(Input->IP_Address));

	strcpy(FileBuffer, TEMPPATH "sambascanner_process_ipaddrXXXXXX");
	errno = 0;
	fd = (int)mkstemp(FileBuffer);
	if (errno != 0) {
		message(1, "Couldn't create temporary file in Process_IP_Address_Thread: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	snprintf(Buffer, sizeof(Buffer), "-A %d.%d.%d.%d", IP_ADDRESS(Input->IP_Address));
	ProcID = Exec_Timeout(Options.NmbLookup_Executable, Buffer, 5.0, 5.0, fd, Input->Processes, Input->NumberProcesses);
	Result = WaitFor(ProcID, Input->Processes);

	Parse_IP_Address_Info(FileBuffer, Input->Hostlist);
	if (!Input->PreserveTemp) unlink(FileBuffer);

	free(Parameters);
	ThreadCnt_Dec();
	pthread_exit(NULL);
	return NULL;
}

